
# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

from prettytable import PrettyTable, SINGLE_BORDER
from os import listdir
from os.path import isfile, join, isdir

def create_table(headers, rows):
    table = PrettyTable()
    table.set_style(SINGLE_BORDER)
    table.field_names = headers
    table.add_rows(rows)
    return table

def read_files_in_directory(directory_path):
    """
    Returns full path to files in directory

    Parameters:
        directory_path: path to directory
    
    Returns:
        list: full path to files in directory
    """
    if isdir(directory_path):
        return [join(directory_path, f) for f in listdir(directory_path) if isfile(join(directory_path, f))]
    else:
        return []



# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

import json
import random
from passlib.hash import sha512_crypt
import string
import os
from pathlib import Path
import click
from bs4 import BeautifulSoup
import requests
import re

import packerpy
import python_terraform
from tectonic.ssh import ssh_version
from tectonic.ansible import Ansible
from tectonic.constants import OS_DATA
from tectonic.utils import create_table

import importlib.resources as tectonic_resources


class DeploymentException(Exception):
    pass


class TerraformRunException(Exception):
    pass


def run_terraform_cmd(t, cmd, variables, **args):
    return_code, stdout, stderr = t.cmd(
        cmd, no_color=python_terraform.IsFlagged, var=variables, **args
    )  # TODO: unused variables stdout
    if return_code != 0:
        error = f"ERROR: terraform {cmd} returned an error: {stderr}"
        raise TerraformRunException(error)
    return stdout


class Deployment:
    """Deployment class."""

    # These will be instantiated by the subclass.
    client = None
    ansible_services_path = tectonic_resources.files('tectonic') / 'services' / 'ansible' / 'configure_services.yml'
    cr_packer_path = tectonic_resources.files('tectonic') / 'image_generation' / 'create_image.pkr.hcl'

    def __init__(
        self,
        description,
        client,
        gitlab_backend_url,
        gitlab_backend_username,
        gitlab_backend_access_token,
        packer_executable_path,
    ):
        """Initialize the deployment object.

        Attributes:
            description: Description object with the CyberRange description.
            client: The client to use.
            gitlab_backend_url: URL of the GitLab backend.
            gitlab_backend_username: Username of the GitLab backend.
            gitlab_backend_access_token: Access token of the GitLab backend.
            packer_executable_path: Path to the Packer executable.
        """
        self.description = description
        self.client = client
        self.gitlab_backend_url = gitlab_backend_url
        self.gitlab_backend_username = gitlab_backend_username
        self.gitlab_backend_access_token = gitlab_backend_access_token
        self.packer_executable_path = packer_executable_path

    def generate_backend_config(self, terraform_dir):
        """
        Generate Gitlab backend configuration to store Terraform states.

        Parameters:
            terraform_dir (str): path to the terraform module.
        """
        tf_mod_name = os.path.basename(os.path.normpath(terraform_dir))
        address = f"{self.gitlab_backend_url}/{self.description.institution}-{self.description.lab_name}-{tf_mod_name}"

        return [
            f"address={address}",
            f"lock_address={address}/lock",
            f"unlock_address={address}/lock",
            f"username={self.gitlab_backend_username}",
            f"password={self.gitlab_backend_access_token}",
            "lock_method=POST",
            "unlock_method=DELETE",
            "retry_wait_min=5",
        ]

    def terraform_apply(self, terraform_dir, variables, resources=None):
        """
        Execute terraform apply command.

        Parameters:
            terraform_dir (str): path to the terraform module.
            variables (dict): variables of the terraform module.
            resources (list(str)): name of terraform resources for target apply.
        """
        t = python_terraform.Terraform(working_dir=terraform_dir)
        run_terraform_cmd(
            t,
            "init",
            [],
            reconfigure=python_terraform.IsFlagged,
            backend_config=self.generate_backend_config(terraform_dir),
        )
        run_terraform_cmd(t, "plan", variables, input=False, target=resources)
        run_terraform_cmd(
            t, "apply", variables, auto_approve=True, input=False, target=resources
        )


    def get_deploy_cr_vars(self):
        """Build the terraform variable dictionary for deployment of the CyberRange."""
        raise NotImplementedError

    def _deploy_cr(self, terraform_module, terraform_variable, instances):
        """
        Deploy cyber range infrastructure.

        Parameters:
            terraform_module (str): Path to the terraform module to use
            terraform_variable (json): json with variables for terraform module.
            instances (list(int)): instances ids. If the list is not empty only the specified instances are created, otherwise all instances are generated.

        """
        resources_to_create = None
        if instances is not None:
            resources_to_create = self.get_cr_resources_to_target_apply(instances)
        self.terraform_apply(
            terraform_module,
            variables=terraform_variable,
            resources=resources_to_create,
        )

    def terraform_destroy(self, terraform_dir, variables, resources=None):
        """
        Execute terraform destroy command.

        Parameters:
            terraform_dir (str): path to the terraform module.
            variables (dict): variables of the terraform module.
            resources (list(str)): name of terraform resources for target destroy.
        """
        t = python_terraform.Terraform(working_dir=terraform_dir)
        run_terraform_cmd(
            t,
            "init",
            [],
            reconfigure=python_terraform.IsFlagged,
            backend_config=self.generate_backend_config(terraform_dir),
        )
        run_terraform_cmd(
            t, "destroy", variables, auto_approve=True, input=False, target=resources
        )

    def _destroy_cr(self, terraform_path, terraform_variables, instances):
        """
        Destroy cyber range infraestructure.

        Parameters:
            terraform_path (str): path to the terraform module
            terraform_variables (json): json with variables for terraform
            instances (list(int)): instances ids. If the list is not empty only the specified instances are destroyed, otherwise all instances are destroyed.
        """
        resources_to_destroy = None
        if instances is not None:  # Target destroy
            resources_to_destroy = self.get_cr_resources_to_target_destroy(instances)
        self.terraform_destroy(
            terraform_path,
            variables=terraform_variables,
            resources=resources_to_destroy,
        )

    def terraform_recreate(self, terraform_dir, machines):
        """
        Recreate instances machines.

        Parameters:
            machines (list(str)): names of the terraform resources associated with the machines to be recreated.
        """
        t = python_terraform.Terraform(working_dir=terraform_dir)
        run_terraform_cmd(
            t,
            "init",
            [],
            reconfigure=python_terraform.IsFlagged,
            backend_config=self.generate_backend_config(terraform_dir),
        )
        run_terraform_cmd(
            t,
            "plan",
            variables=self.get_deploy_cr_vars(),
            input=False,
            replace=machines,
        )
        run_terraform_cmd(
            t,
            "apply",
            variables=self.get_deploy_cr_vars(),
            auto_approve=True,
            input=False,
            replace=machines,
        )

    def _create_packer_images(self, packer_path, variables):
        """
        Build images using Packer.

        Parameters:
            packer_path (str): path to the Packer module.
            variables (dict): variables of the Packer module.
        """
        p = packerpy.PackerExecutable(executable_path=self.packer_executable_path)
        return_code, stdout, stderr = p.execute_cmd("init", str(packer_path))
        if return_code != 0:
            raise TerraformRunException(
                f"ERROR: packer init returned an error:\n{stdout.decode()}"
            )
        return_code, stdout, stderr = p.build(str(packer_path), var=variables)
        # return_code, stdout, stderr = p.build(packer_path, var=variables, on_error="abort")
        if return_code != 0:
            raise TerraformRunException(
                f"ERROR: packer build returned an error:\n{stdout.decode()}"
            )

    def create_cr_images(self, guests=None):
        """
        Create scenario base images.

        Parameters:
            guests (list(str)): guests names. If the list is not empty only the base images of the specified guests are generated, otherwise all the base images are generated.
        """
        machines = {}
        for guest_name in guests or self.description.guest_settings.keys():
            monitor = True if self.description.deploy_elastic and self.description.get_guest_attr(guest_name, "monitor", False) and self.description.monitor_type == "endpoint" else False
            machines[guest_name] = {
                "base_os": self.description.get_guest_attr(
                    guest_name, "base_os", self.description.default_os
                ),
                "endpoint_monitoring" : monitor,
            }
            if self.description.platform == "libvirt":
                machines[guest_name]["vcpu"] = self.description.get_guest_attr(guest_name, "vcpu", 1)
                machines[guest_name]["memory"] = self.description.get_guest_attr(guest_name, "memory", 1024)
                machines[guest_name]["disk"] = self.description.get_guest_attr(guest_name, "disk", 10)
            elif self.description.platform == "aws":
                machines[guest_name]["disk"] = self.description.get_guest_attr(guest_name, "disk", 8)
                machines[guest_name]["instance_type"] = self.description.instance_type.get_guest_instance_type(
                    self.description.get_guest_attr(guest_name, "memory", 1),
                    self.description.get_guest_attr(guest_name, "vcpu", 1),
                    monitor,
                    self.description.monitor_type,
                )
        args = {
            "ansible_playbooks_path": self.description.ansible_playbooks_path,
            "ansible_scp_extra_args": "'-O'" if ssh_version() >= 9 else "",
            "ansible_ssh_common_args": self.description.ansible_ssh_common_args,
            "aws_region": self.description.aws_region,
            "instance_number": self.description.instance_number,
            "institution": self.description.institution,
            "lab_name": self.description.lab_name,
            "libvirt_proxy": self.description.libvirt_proxy,
            "libvirt_storage_pool": self.description.libvirt_storage_pool,
            "libvirt_uri": self.description.libvirt_uri,
            "machines_json": json.dumps(machines),
            "os_data_json": json.dumps(OS_DATA),
            "platform": self.description.platform,
            "remove_ansible_logs": str(not self.description.keep_ansible_logs),
            "elastic_version": self.description.elastic_stack_version
        }
        self._create_packer_images(self.cr_packer_path, args)

    def get_teacher_access_username(self):
        """
        Get username for connection to teacher access host.
        """
        return OS_DATA[self.description.default_os]["username"]

    def get_teacher_access_ip(self):
        """
        Returns the public IP assigned to the teacher access host.
        """
        if self.description.teacher_access == "host":
            return self.client.get_machine_public_ip(
                self.description.get_teacher_access_name()
            )
        return None

    def get_student_access_users(self):
        """Returns a dictionary of users with username, password, password_hash and authorized_keys.
        """
        users = {}
        random.seed(self.description.random_seed)
        for i in self.description.get_instance_range():
            username = f"{self.description.student_prefix}{i:02d}"
            users[username] = {}
            if self.description.create_student_passwords:
                (password, salt) = self._generate_password()
                users[username]["password"] = password
                users[username]["password_hash"] = sha512_crypt.using(salt=salt).hash(
                    password
                )

            if self.description.student_pubkey_dir:
                users[username]["authorized_keys"] = self.description.read_pubkeys(
                    os.path.join(self.description.student_pubkey_dir, username)
                )

        return users
        

    def _student_access(self, instances, guests):
        """Creates users for the students in all GUESTS of INSTANCES.

        Generates pseudo-random passwords and/or sets public SSH keys for the users.
        Returns a dictionary of created users.
        """
        playbook = tectonic_resources.files('tectonic') / 'playbooks' / 'trainees.yml'
        
        users = self.get_student_access_users()
        only_instances = True
        if guests and "student_access" in guests:
            only_instances = False
        ansible = Ansible(self)
        ansible.run(
            instances=instances,
            guests=guests,
            copies=None,  # Apply playbook to all copies
            playbook=playbook,
            only_instances=only_instances,
            extra_vars={"users": users, "prefix": self.description.student_prefix, "ssh_password_login": self.description.create_student_passwords},
            quiet=True,
        )
        return users

    def student_access(self, instances):
        """Creates users for the students in all entry points of INSTANCES.

        Generates pseudo-random passwords and/or sets public SSH keys for the users.
        Returns a dictionary of created users.
        """
        entry_points = [ base_name for base_name, guest in self.description.guest_settings.items()
                         if guest and guest.get("entry_point") ]
        self._student_access(instances, entry_points)


    def _generate_password(self):
        """Generate a pseudo random password and salt."""
        characters = string.ascii_letters + string.digits
        password = "".join(random.choice(characters) for _ in range(12))
        salt = "".join(random.choice(characters) for _ in range(16))
        return password, salt

    # Abstract methods to be implemented by platform dependent sub-classes:

    def delete_cr_images(self, guests=None):
        """Delete guests scenario images."""
        raise NotImplementedError

    def get_instance_status(self, machine):
        """
        Returns a dictionary with the instance status of machine ID.

        Parameters:
            machine (str): machine identifier

        Returns:
            dict: instance status information
        """
        raise NotImplementedError

    def get_cyberrange_data(self):
        """Returns a dictionary with Cyber Range data."""
        raise NotImplementedError

    def connect_to_instance(self, instance_name, username):
        """
        Interactively connects to a running instance.

        Parameters:
            instance_name (str): name for the instance.
            username: (str): username
        """
        raise NotImplementedError

    def get_ssh_proxy_command(self):
        """Returns the appropriate SSH proxy configuration to access guest machines."""
        return None

    def get_ssh_hostname(self, machine):
        """Returns the appropriate SSH hostname to access guest machines."""
        return self.client.get_machine_private_ip(machine)

    def start_instance(self, instance_name):
        """
        Starts a stopped instance.

        Parameters:
            instance_name (str): name of the instance.
        """
        raise NotImplementedError

    def stop_instance(self, instance_name):
        """
        Stops a running instance.

        Parameters:
            instance_name (str): name of the instance.
        """
        raise NotImplementedError

    def reboot_instance(self, instance_name):
        """
        Reboots a running instance.

        Parameters:
            instance_name (str): name of the instance.
        """
        raise NotImplementedError

    def can_delete_image(self, image_name):
        """
        Return true if the image is not being used by any machine.

        Parameters:
            image_name (str): name of the image.

        Returns:
            bool: true if the image is not being used by any machine, false otherwise.
        """
        raise NotImplementedError

    def get_cr_resources_to_target_apply(self, instances):
        """
        Returns the name of the resources to target apply based on the instance numbers.

        Parameters:
          instances (list(int)): instances to use.

        Returns:
          list(str): resources name to target apply.
        """
        raise NotImplementedError

    def get_cr_resources_to_target_destroy(self, instances):
        """
        Returns the name of the resource to target destroy based on the instance numbers.

        Parameters:
          instances (list(int)): instances to use.

        Returns:
          list(str): resources name to target destroy.
        """
        raise NotImplementedError

    def get_resources_to_recreate(self, instances, guests, copies):
        """
        Returns the name of the terraform resources to recreate based on the machines names.

        Parameters:
          instances (list(int)): instances to use.

        Returns:
          list(str): resource names to recreate.
        """
        raise NotImplementedError

    def get_elastic_latest_version(self):
        """
        Return latest version of elastic stack available

        Returns:
          str: elastic stack version
        """
        try:
            elastic_url = 'https://www.elastic.co/guide/en/elasticsearch/reference/current/es-release-notes.html'
            html_text = requests.get(elastic_url).text
            soup = BeautifulSoup(html_text, 'html.parser')
            versions = soup.find_all('a', attrs={"title": re.compile("Elasticsearch version \\d+\\.\\d+\\.\\d+")})
            latest_version = versions[0].get("title").split(" ")[2]
            return latest_version
        except Exception as e:
            raise DeploymentException(e)

    def get_guest_instance_type(self, base_name):
        """
        Return the guest instance type.

        Parameters:
            base_name (str): guests base name
        """
        return None

    def list_instances(self, instances, guests, copies):
        """
        Get info about Cyber range

        Returns table containing information about cyber range
        """
        raise NotImplementedError
    
    def shutdown(self, instances, guests, copies, stop_services):
        """
        Shutdown cyber range machines.
        """
        raise NotImplementedError
    
    def start(self, instances, guests, copies, start_services):
        """
        Start cyber range machines.
        """
        raise NotImplementedError
    
    def reboot(self, instances, guests, copies, reboot_services):
        """
        Reboot cyber range machines.
        """
        raise NotImplementedError
    
    def recreate(self, instances, guests, copies, recreate_services=False): #TODO: add support for recreating services
        """
        Recreate cyber range machines.
        """
        raise NotImplementedError
    
    def deploy_infraestructure(self, instances):
        """
        Deploy cyber range infraestructure
        """
        raise NotImplementedError

    def destroy_infraestructure(self, instances):
        """
        Destroy cyber range infraestructure
        """
        raise NotImplementedError
    
    def get_elastic_agent_status(self):
        """
        Get elastic agent status
        """
        raise NotImplementedError
    
    def get_parameters(self, instances, directory):
        self.description.parse_machines(instances)
        parameters = self.description.get_parameters(instances)
        headers = ["Instance","Parameters"]
        rows = []
        for parameter in parameters:
            rows.append([parameter,parameters[parameter]])
        if directory:
            if Path(directory).is_dir():
                try:
                    filename = os.path.join(directory, f"{self.description.institution}-{self.description.lab_name}-parameters.json")
                    with open(filename, "w") as parameters_file:
                        parameters_file.write(json.dumps(parameters, indent=4))
                        parameters_file.write("\n")
                        
                        print(f"File {filename} created")
                except Exception as exception:
                    raise TerraformRunException(f"{exception}")
            else:
                raise TerraformRunException(f"ERROR: directory {directory} does not exist.")
        else:
            click.echo(create_table(headers,rows))

    def create_services_images(self, services):
        """
        Create services base images.

        Parameters:
            services (dict(bool)): services images to create.
        """
        machines = {}
        for service in services:
            if services[service]:
                machines[service] = {
                    "base_os": self.description.get_service_base_os(service),
                    "ansible_playbook": str(tectonic_resources.files('tectonic') / 'services' / service / 'base_config.yml'),
                }
                if self.description.platform == "libvirt":
                        machines[service]["vcpu"] = self.description.services[service]["vcpu"]
                        machines[service]["memory"] = self.description.services[service]["memory"]
                        machines[service]["disk"] = self.description.services[service]["disk"]
                elif self.description.platform == "aws":
                    machines[service]["disk"] = self.description.services[service]["disk"]
                    if service in ["caldera", "elastic"]:
                        machines[service]["instance_type"] = self.description.instance_type.get_guest_instance_type(self.description.services[service]["memory"], self.description.services[service]["vcpu"], False, self.description.monitor_type)
                    elif service == "packetbeat":
                        machines[service]["instance_type"] = "t2.micro"
        args = {
            "ansible_scp_extra_args": "'-O'" if ssh_version() >= 9 else "",
            "ansible_ssh_common_args": self.description.ansible_ssh_common_args,
            "aws_region": self.description.aws_region,
            "libvirt_proxy": self.description.libvirt_proxy,
            "libvirt_storage_pool": self.description.libvirt_storage_pool,
            "libvirt_uri": self.description.libvirt_uri,
            "machines_json": json.dumps(machines),
            "os_data_json": json.dumps(OS_DATA),
            "platform": self.description.platform,
            "remove_ansible_logs": str(not self.description.keep_ansible_logs),
            #TODO: pass variables as a json as part of each host
            "elastic_version": self.description.elastic_stack_version, 
            "elastic_latest_version": "yes" if self.description.is_elastic_stack_latest_version else "no",
            "caldera_version": "master",
            "packetbeat_vlan_id": self.description.packetbeat_vlan_id,
        }
        self._create_packer_images(tectonic_resources.files('tectonic') / 'services' / 'image_generation' / 'create_image.pkr.hcl', args)

    def delete_services_images(self, services):
        """Delete services images."""
        raise NotImplementedError

    def get_deploy_services_vars(self):
        """Build the terraform variable dictionary for deployment of the CyberRange services."""
        raise NotImplementedError

    def get_services_status(self):
        """Get services status"""
        raise NotImplementedError
    
    def _elastic_install_endpoint(self, instances=None):
        """
        Install elastic endpoint on instances

        Parameters:
            instances(list(int)): instances number
        """
        elastic_name = self.description.get_service_name("elastic")
        if self.get_instance_status(elastic_name) == "RUNNING":
            elastic_ip = self.get_ssh_hostname(elastic_name)
            playbook = tectonic_resources.files('tectonic') / 'services' / 'elastic' / 'get_info.yml'
            result = self._get_service_info("elastic",playbook,{"action":"get_token_by_policy_name","policy_name":self.description.endpoint_policy_name})
            endpoint_token = result[0]["token"]
            extra_vars = {
                "institution": self.description.institution,
                "lab_name": self.description.lab_name,
                "token": endpoint_token,
                "elastic_url": f"https://{elastic_ip}:8220",
            }
            ansible = Ansible(deployment=self)
            guests_to_monitor = self.description.get_machines_to_monitor()
            machines = self.description.parse_machines(instances=instances, guests=guests_to_monitor)
            inventory = ansible.build_inventory(machine_list=machines, extra_vars=extra_vars)
            ansible.wait_for_connections(inventory=inventory)
            ansible.run(inventory=inventory,playbook=tectonic_resources.files('tectonic') / 'services' / 'elastic' / 'endpoint_install.yml', quiet=True)
        else:
            raise DeploymentException("Elastic machine is not running. Unable to install endpoints.")

    def _caldera_install_agent(self, instances=None):
        """
        Install caldera agent on instances

        Parameters:
            instances(list(int)): instances number
        
        """
        extra_vars = {
            "institution": self.description.institution,
            "lab_name": self.description.lab_name,
            "caldera_ip": self.get_ssh_hostname(self.description.get_service_name("caldera")),
            "caldera_agent_type": "red",
        }
        ansible = Ansible(deployment=self)
        red_team_machines = self.description.get_red_team_machines()
        if len(red_team_machines) > 0:
            machines_red = self.description.parse_machines(instances=instances, guests=red_team_machines)
            inventory_red = ansible.build_inventory(machine_list=machines_red, extra_vars=extra_vars)
            ansible.wait_for_connections(inventory=inventory_red)
            ansible.run(inventory = inventory_red,
                        playbook = tectonic_resources.files('tectonic') / 'services' / 'caldera' / 'agent_install.yml',
                        quiet = True)

        extra_vars["caldera_agent_type"] = "blue"
        blue_team_machines = self.description.get_blue_team_machines()
        if len(blue_team_machines) > 0:
            machines_blue = self.description.parse_machines(instances=instances, guests=blue_team_machines)
            inventory_blue = ansible.build_inventory(machine_list=machines_blue, extra_vars=extra_vars)
            ansible.wait_for_connections(inventory=inventory_blue)
            ansible.run(inventory = inventory_blue,
                        playbook = tectonic_resources.files('tectonic') / 'services' / 'caldera' / 'agent_install.yml',
                        quiet = True)

    def _get_services_guest_data(self):
        """
        Build the terraform variable dictionary for deployment of the CyberRange services.
        """
        raise NotImplementedError
    
    def _get_services_network_data(self):
        """Compute the complete list of services subnetworks."""
        raise NotImplementedError
    
    def _get_service_password(self, service_base_name):
        """
        Get service credentials. Use Ansible to connect to machine and get the credentials.

        Parameters:
            service_base_name (str): service name (example: caldera).
        """
        try:
            ansible = Ansible(self)
            ansible.run(
                instances=None,
                guests=[service_base_name],
                playbook = tectonic_resources.files('tectonic') / 'playbooks' / 'services_get_password.yml',
                only_instances=False,
                username=OS_DATA[self.description.get_service_base_os(service_base_name)]["username"],
                quiet=True
            )
            credentials = ansible.debug_outputs[0]["password.stdout"].split("\n")
            result = {}
            for credential in credentials:
                credential_split = credential.split(" ")
                result[credential_split[0]] = credential_split[1]
            return result
        except Exception as e:
            raise TerraformRunException(f"{e}")
        
    def _destroy_services(self, terraform_path, terraform_variables, instances):
        """
        Destroy cyber range services infraestructure.

        Parameters:
            terraform_path (str): path to the terraform module
            terraform_variables (json): json with variables for terraform
            instances (list(int)): instances ids. If the list is not empty only the specified instances are destroyed, otherwise all instances are destroyed.
        """
        resources_to_destroy = None
        if instances is not None:  # Target destroy
            resources_to_destroy = self.get_services_resources_to_target_destroy(instances)
        self.terraform_destroy(terraform_path, variables=terraform_variables, resources=resources_to_destroy)

    def get_services_resources_to_target_destroy(self, instances):
        """
        Returns the name of the services resources to target destroy based on the instance numbers.

        Parameters:
          instances (list(int)): instances to use.

        Returns:
          list(str): resources name to target destroy.
        """
        raise NotImplementedError
    
    def _deploy_services(self, terraform_module, terraform_variable, instances):
        """
        Deploy cyber range infrastructure.

        Parameters:
            terraform_module (str): Path to the terraform module to use
            terraform_variable (json): json with variables for terraform module.
            instances (list(int)): instances ids. If the list is not empty only the specified instances are created, otherwise all instances are generated.

        """
        resources_to_create = None
        if instances is not None:
            resources_to_create = self.get_services_resources_to_target_apply(instances)
        self.terraform_apply(terraform_module, variables=terraform_variable, resources=resources_to_create)

    def get_sevices_resources_to_target_apply(self, instances):
        """
        Returns the name of the services resources to target apply based on the instance numbers.

        Parameters:
          instances (list(int)): instances to use.

        Returns:
          list(str): resources name to target apply.
        """
        raise NotImplementedError
    
    def _get_service_info(self, service_base_name, playbook, ansible_vars):
        """
        Get service info. Use Ansible to execute action against service and get specific info.

        Parameters:
            service_base_name (str): service name (example: caldera).
            playbook (Path): Ansible playbook to apply.
            ansible_vars (dict): Ansible variables to use.
        """
        try:
            ansible = Ansible(self)
            ansible.run(
                instances=None,
                guests=[service_base_name],
                playbook=playbook,
                only_instances=False,
                username=OS_DATA[self.description.get_service_base_os(service_base_name)]["username"],
                quiet=True,
                extra_vars=ansible_vars,
            )
            return ansible.debug_outputs
        except Exception as e:
            raise TerraformRunException(f"{e}")




# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

import logging
import pprint
from pathlib import Path
import ansible_runner
import importlib.resources as tectonic_resources

logger = logging.getLogger(__name__)

class AnsibleException(Exception):
    """ AnsibleException class """
    pass


class Ansible:
    """ Class for managing Ansible connections. """

    def __init__(self, deployment):
        self.output = ""
        self.debug_outputs = []
        self.deployment = deployment

    def _ansible_callback(self, event_data):
        if event_data['stdout']:
            self.output += f"\n{event_data['stdout']}"
            event_data = event_data.get("event_data",None)
            if event_data:
                resolved_action = event_data.get("resolved_action",None)
                if resolved_action == "ansible.builtin.debug" and event_data.get("res",None):
                    debug_output = event_data["res"]
                    if debug_output.get("_ansible_verbose_always",None):
                        del debug_output["_ansible_verbose_always"]
                    if debug_output.get("_ansible_no_log",None):
                        del debug_output["_ansible_no_log"]
                    if debug_output.get("changed",None):
                        del debug_output["changed"]
                    self.debug_outputs.append(debug_output)
        return True

    def build_inventory(self, machine_list=None, username=None, extra_vars=None):
        if extra_vars is None:
            extra_vars = {}
        if machine_list is None:
            machine_list = []
        inventory = {}

        deployment = self.deployment
        description = deployment.description
        parameters = deployment.description.get_parameters()

        ssh_args = description.ansible_ssh_common_args

        proxy_command = deployment.get_ssh_proxy_command()
        if proxy_command:
            ssh_args += f' -o ProxyCommand="{proxy_command}"'

        for machine in machine_list:
            base_name = description.get_base_name(machine)
            ansible_username = username or description.get_guest_username(base_name)
            hostname = deployment.get_ssh_hostname(machine)

            if not inventory.get(base_name):
                inventory[base_name] = {
                    "hosts": {},
                    "vars": {
                                "ansible_become": True,
                                "basename": base_name,
                                "instances": description.instance_number,
                                "platform": description.platform,
                                "institution": description.institution,
                                "lab_name": description.lab_name
                            } | extra_vars,
                }

            inventory[base_name]["hosts"][machine] = {
                "ansible_host": hostname,
                "ansible_user": ansible_username,
                "ansible_ssh_common_args": ssh_args,
                "instance": description.get_instance_number(machine),
                "copy": description.get_copy(machine),
                "parameter": parameters[description.get_instance_number(machine)] if description.get_instance_number(machine) else {}
            }

            if description.get_guest_attr(
                    base_name, "base_os", description.default_os
            ) == "windows_srv_2022":
                inventory[base_name]["hosts"][machine]["ansible_shell_type"] = "powershell"
                inventory[base_name]["hosts"][machine]["ansible_become_method"] = "runas"
                inventory[base_name]["hosts"][machine]["ansible_become_user"] = \
                    description.get_guest_username(base_name)
            else:
                # Make sure we load environment variables within sudo shell
                inventory[base_name]["hosts"][machine]["become_flags"] = "-i"
        return inventory

    def run(self,
            instances=None,
            guests=None,
            copies=None,
            only_instances=True,
            username=None,
            inventory=None,
            playbook=None,
            extra_vars=None,
            quiet=False,
            verbosity=0,
            exclude=[]
            ):
        """ Run Ansible playbook.

        Args:
            instances (list): List of instances to run the playbook on.
            guests (list): List of guests to run the playbook on.
            copies (list(int)): List of copies to run the playbook on.
            only_instances (bool): If True, only run the playbook on instances.
            username (str): Username to use for Ansible connections.
            inventory (dict): Ansible inventory dictionary.
            playbook (str): Ansible playbook to run.
            extra_vars (dict): Extra variables to pass to Ansible.
            quiet (bool): If True, suppress ansible output, and raise an exception on error.
            verbosity (int): Ansible verbosity level.
            exclude (list): guest name of machines to exclude from applying playbook

        Raises:
            AnsibleException: If the playbook fails and quiet is False.
        """
        default_playbook = (Path(self.deployment.description.ansible_playbooks_path).
                            joinpath("after_clone.yml"))

        if not playbook:
            if default_playbook.is_file():
                playbook = default_playbook.resolve().as_posix()
            else:
                if not quiet:
                    print("No playbook to run")
                return
        else:
            playbook = Path(playbook).resolve().as_posix()
        if not inventory:
            machine_list = self.deployment.description.parse_machines(
                instances, guests, copies, only_instances, exclude
            )
            inventory = self.build_inventory(
                machine_list=machine_list,
                username=username,
                extra_vars=extra_vars
            )
        self.output = ""
        self.debug_outputs = []
        extravars = {"ansible_no_target_syslog" : not self.deployment.description.keep_ansible_logs }
        r = ansible_runner.interface.run(
            inventory=inventory,
            playbook=playbook,
            quiet=quiet,
            verbosity=verbosity,
            event_handler=self._ansible_callback,
            extravars=extravars
        )
        logger.info(self.output)

        if (r.rc != 0 or r.status != "successful") and quiet:
            raise AnsibleException(self.output)

    def wait_for_connections(self, instances=None, guests=None, copies=None, only_instances=True, exclude=[], username=None, inventory=None):
        """Wait for machines to respond to ssh connections for ansible."""

        playbook = tectonic_resources.files('tectonic') / 'playbooks' / 'wait_for_connection.yml'
        return self.run(
            instances=instances, guests=guests, copies=copies,
            only_instances=only_instances,
            playbook=playbook, quiet=True,
            exclude=exclude, username=username, inventory=inventory
        )




# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

import json
import click
import ipaddress
import datetime

from tectonic.aws import Client
from tectonic.constants import *
from tectonic.deployment import Deployment
from tectonic.ssh import interactive_shell
from tectonic.utils import create_table
from tectonic.ansible import Ansible

import importlib.resources as tectonic_resources
    


class DeploymentAWSException(Exception):
    pass


class AWSDeployment(Deployment):
    """Deployment class for AWS."""

    EIC_ENDPOINT_SSH_PROXY = "aws ec2-instance-connect open-tunnel --instance-id %h"

    def __init__(
        self,
        description,
        gitlab_backend_url,
        gitlab_backend_username,
        gitlab_backend_access_token,
        packer_executable_path,
    ):
        super().__init__(
            description,
            Client(description),
            gitlab_backend_url,
            gitlab_backend_username,
            gitlab_backend_access_token,
            packer_executable_path,
        )

    def generate_backend_config(self, terraform_dir):
        return super().generate_backend_config(terraform_dir)

    def get_deploy_cr_vars(self):
        """Build the terraform variable dictionary for deployment of the CyberRange."""
        return {
            "institution": self.description.institution,
            "lab_name": self.description.lab_name,
            "instance_number": self.description.instance_number,
            "aws_region": self.description.aws_region,
            "network_cidr_block": self.description.network_cidr_block,
            "services_network_cidr_block": self.description.services_network,
            "internet_network_cidr_block": self.description.internet_network,
            "ssh_public_key_file": self.description.ssh_public_key_file,
            "teacher_access": self.description.teacher_access,
            "authorized_keys": self.description.authorized_keys,
            "subnets_json": json.dumps(self.description.subnets),
            "guest_data_json": json.dumps(self.description.get_guest_data()),
            "aws_default_instance_type": self.description.aws_default_instance_type,
            "default_os": self.description.default_os,
            "os_data_json": json.dumps(OS_DATA),
            "configure_dns": self.description.configure_dns,
            "services_internet_access": self.description.deploy_elastic,
            "monitor_type": self.description.monitor_type,
        }

    def delete_cr_images(self, guests=None):
        guests = guests or self.description.guest_settings.keys()
        for guest_name in guests:
            image_name = self.description.get_image_name(guest_name)
            if not self.can_delete_image(image_name):
                raise DeploymentAWSException(
                    f"Unable to delete image {image_name} because it is being used."
                )
        for guest_name in guests:
            self.client.delete_image(self.description.get_image_name(guest_name))

    def get_instance_status(self, machine):
        """Returns a dictionary with the instance status of machine ID."""
        return self.client.get_instance_status(machine)

    def get_cyberrange_data(self):
        """Get information about cyber range"""
        headers = ["Name", "Value"]
        rows = [["Student Access IP", self.client.get_machine_public_ip(f"{self.description.institution}-{self.description.lab_name}-student_access")]]
        if self.description.teacher_access == "host":
            rows.append(["Teacher Access IP", self.client.get_machine_public_ip(f"{self.description.institution}-{self.description.lab_name}-teacher_access")])
        if len(self.description.get_services_to_deploy()) > 0:
            rows.append(["",""])
            if self.description.deploy_elastic:
                if self.get_instance_status(self.description.get_service_name("elastic")) == "RUNNING":
                    elastic_ip = self.get_ssh_hostname(self.description.get_service_name("elastic"))
                    elastic_credentials = self._get_service_password("elastic")
                    rows.append(["Kibana URL", f"https://{elastic_ip}:5601"])
                    rows.append(["Kibana user (username: password)", f"elastic: {elastic_credentials['elastic']}"])
                    if self.description.deploy_caldera:
                        rows.append(["",""])
                else:
                    return "Unable to get Elastic info right now. Please make sure de Elastic machine is running."
            if self.description.deploy_caldera:
                if self.get_instance_status(self.description.get_service_name("caldera")) == "RUNNING":
                    caldera_ip = self.get_ssh_hostname(self.description.get_service_name("caldera"))
                    caldera_credentials = self._get_service_password("caldera")
                    rows.append(["Caldera URL", f"https://{caldera_ip}:8443"])
                    rows.append(["Caldera user (username: password)", f"red: {caldera_credentials['red']}"])
                    rows.append(["Caldera user (username: password)", f"blue: {caldera_credentials['blue']}"])
                else:
                    return "Unable to get Caldera info right now. Please make sure de Caldera machine is running."
        return create_table(headers,rows)

    def connect_to_instance(self, instance_name, username):
        if instance_name == self.description.get_teacher_access_name():
            interactive_shell(
                self.get_teacher_access_ip(), self.get_teacher_access_username()
            )
            return

        username = username or self.description.get_guest_username(
            self.description.get_base_name(instance_name)
        )
        if self.description.teacher_access == "host":
            hostname = self.client.get_machine_private_ip(instance_name)
            gateway = (self.get_teacher_access_ip(), self.get_teacher_access_username())
        else:
            hostname = self.client.get_instance_property(instance_name, "InstanceId")
            gateway = self.EIC_ENDPOINT_SSH_PROXY

        if not hostname:
            raise DeploymentAWSException(f"Instance {instance_name} not found.")

        interactive_shell(hostname, username, gateway)

    def get_ssh_proxy_command(self):
        if self.description.teacher_access == "endpoint":
            proxy_command = self.EIC_ENDPOINT_SSH_PROXY
        else:
            access_ip = self.get_teacher_access_ip()
            username = self.get_teacher_access_username()
            connection_string = f"{username}@{access_ip}"
            proxy_command = f"ssh {self.description.ansible_ssh_common_args} -W %h:%p {connection_string}"

        return proxy_command

    def get_ssh_hostname(self, machine):
        if self.description.teacher_access == "endpoint":
            return self.client.get_instance_property(machine, "InstanceId")
        else:
            return self.client.get_machine_private_ip(machine)

    def start_instance(self, instance_name):
        self.client.start_instance(instance_name)

    def stop_instance(self, instance_name):
        self.client.stop_instance(instance_name)

    def reboot_instance(self, instance_name):
        self.client.reboot_instance(instance_name)

    def _get_machines_resources_name(self, instances):
        """
        Returns the name of the aws_instance resource of the gsi_lab_aws module for the instances

        Parameters:
          instances (list(int)): instances to use

        Returns:
          list(str): resources name of the aws_instances for the instances
        """
        machines = self.description.parse_machines(instances, None, None, False, self.description.get_services_to_deploy())
        resources = []
        for machine in machines:
            resources.append('aws_instance.machines["' f"{machine}" '"]')
        return resources

    def _get_route_table_resources_name(self, instances):
        """
        Returns the name of the aws_route_table resource of the gsi_lab_aws module for the instances

        Parameters:
          instances (list(int)): instances to use

        Returns:
          list(str): resources name of the aws_route_table for the instances
        """
        resources = []
        for instance in filter(
            lambda i: i <= self.description.instance_number,
            instances or range(1, self.description.instance_number + 1),
        ):
            for network in self.description.topology:
                resources.append(
                    'aws_route_table_association.scenario_internet_access["'
                    f"{self.description.institution}-{self.description.lab_name}-{str(instance)}-{network['name']}"
                    '"]'
                )
        return resources

    def _get_dns_resources_name(self, instances):
        """
        Returns the name of the aws dns resource of the gsi_lab_aws module for the instances

        Parameters:
          instances (list(str)): instances to use

        Returns:
          list(str): resources name of the aws dns resources for the instances
        """
        resources = ["aws_route53_zone.reverse[0]"]
        for network in self.description.topology:
            resources.append(
                'aws_route53_zone.zones["{name}"]'.format(name=network["name"])
            )
        for instance in filter(
            lambda i: i <= self.description.instance_number,
            instances or range(1, self.description.instance_number + 1),
        ):
            for guest in self.description.guest_settings.keys():
                for network in self.description.get_guest_networks(guest):
                    if self.description.get_guest_copies(guest) == 1:
                        resources.append(
                            'aws_route53_record.records["'
                            f"{guest}-{str(instance)}-{network}"
                            '"]'
                        )
                        resources.append(
                            'aws_route53_record.records_reverse["'
                            f"{guest}-{str(instance)}-{network}"
                            '"]'
                        )
                    else:
                        for copy in self.description.get_copy_range(guest):
                            resources.append(
                                'aws_route53_record.records["'
                                f"{guest}-{str(copy)}-{str(instance)}-{network}"
                                '"]'
                            )
                            resources.append(
                                'aws_route53_record.records_reverse["'
                                f"{guest}-{str(copy)}-{str(instance)}-{network}"
                                '"]'
                            )
        return resources
    
    def _get_services_dns_resources_name(self):
        """
        Returns the name of the aws dns resource of the services-aws module

        Returns:
          list(str): resources name of the aws dns resources for the services
        """
        resources = []
        for network in self._get_services_network_data():
            resources.append('aws_route53_zone.zones["'f"{network.split('-')[2]}"'"]')
            services_data = self._get_services_guest_data() 
            for service in services_data:
                interfaces_data = services_data[service]["interfaces"]
                for interface in interfaces_data:
                    resources.append('aws_route53_record.records["'f"{service.split('-')[2]}-{interfaces_data[interface]['network_name']}"'"]')
                    resources.append('aws_route53_record.records_reverse["'f"{service.split('-')[2]}-{interfaces_data[interface]['network_name']}"'"]')
        return resources

    def _get_security_group_resources_name(self, instances):
        """
        Returns the name of the aws_security_group resource of the gsi_lab_aws module for the instances

        Parameters:
          instances (list(str)): instances to use

        Returns:
          list(str): resources name of the aws_security_group for the instances
        """
        resources = []
        for instance in filter(
            lambda i: i <= self.description.instance_number,
            instances or range(1, self.description.instance_number + 1),
        ):
            for network in self.description.topology:
                resources.append(
                    'aws_security_group.subnet_sg["'
                    f"{self.description.institution}-{self.description.lab_name}-{str(instance)}-{network['name']}"
                    '"]'
                )
        return resources
    
    def _get_services_security_group_resources_name(self):
        """
        Returns the name of the aws_security_group resource of the services-aws module for the instances

        Returns:
          list(str): resources name of the subnetworks security groups for the services-aws module
        """
        resources = []
        for network in self._get_services_network_data():
            resources.append('aws_security_group.subnet_sg["'f"{network}"'"]')
        return resources

    def _get_subnet_resources_name(self, instances):
        """
        Returns the name of the aws_subnet resource of the gsi_lab_aws module for the instances

        Parameters:
          instances (list(str)): instances to use

        Returns:
          list(str): resources name of the aws_subnet for the instances
        """
        resources = []
        for instance in filter(
            lambda i: i <= self.description.instance_number,
            instances or range(1, self.description.instance_number + 1),
        ):
            for network in self.description.topology:
                resources.append(
                    'aws_subnet.instance_subnets["'
                    f"{self.description.institution}-{self.description.lab_name}-{str(instance)}-{network['name']}"
                    '"]'
                )
        return resources

    def _get_interface_resources_name(self, instances):
        """
        Returns the name of the aws_network_interface resource of the gsi_lab_aws module for the instances

        Parameters:
          instances (list(str)): instances to use

        Returns:
          list(str): resources name of the aws_network_interface for the instances
        """
        resources = []
        for instance in filter(
            lambda i: i <= self.description.instance_number,
            instances or range(1, self.description.instance_number + 1),
        ):
            for guest in self.description.guest_settings.keys():
                network_index = 1
                for _ in self.description.get_guest_networks(guest):
                    if self.description.get_guest_copies(guest) == 1:
                        resources.append(
                            'aws_network_interface.interfaces["'
                            f"{self.description.institution}-{self.description.lab_name}-{instance}-{guest}-{network_index}"
                            '"]'
                        )
                    else:
                        for copy in self.description.get_copy_range(guest):
                            resources.append(
                                'aws_network_interface.interfaces["'
                                f"{self.description.institution}-{self.description.lab_name}-{instance}-{guest}-{copy}-{network_index}"
                                '"]'
                            )
                    network_index = network_index + 1
        return resources
    
    def _get_services_interface_resources_name(self):
        """
        Returns the name of the aws_network_interface resource of the services-aws module


        Returns:
          list(str): resources name of the aws_network_interface for the services
        """
        resources = []
        services_data = self._get_services_guest_data() 
        for service in services_data:
            for interface in services_data[service]["interfaces"]:
                resources.append('aws_network_interface.interfaces["'f"{interface}"'"]')
        return resources
    
    def _get_services_instances_name(self):
        """
        Returns the name of the aws_instance resource of the services-aws module

        Returns:
          list(str): resources name of the aws_instances for the services
        """
        resources = []
        for service in self._get_services_guest_data():
                resources.append('aws_instance.machines["'f"{service}"'"]')
        return resources

    def _get_sessions_resources_name(self, instances):
        """
        Returns the name of the aws_ec2_traffic_mirror_session resource of the services-aws module for the instances

        Parameters:
          instances (list(str)): instances to use

        Returns:
          list(str): resources name of the aws_ec2_traffic_mirror_session for the instances
        """
        resources = []
        for instance in filter(
            lambda i: i <= self.description.instance_number,
            instances or range(1, self.description.instance_number + 1),
        ):
            for guest in self.description.get_machines_to_monitor():
                network_index = 1
                for _ in self.description.get_guest_networks(guest):
                    if self.description.get_guest_copies(guest) == 1:
                        resources.append(
                            'aws_ec2_traffic_mirror_session.session["'
                            f"{self.description.institution}-{self.description.lab_name}-{instance}-{guest}-{network_index}"
                            '"]'
                        )
                    else:
                        for copy in self.description.get_copy_range(guest):
                            resources.append(
                                'aws_ec2_traffic_mirror_session.session["'
                                f"{self.description.institution}-{self.description.lab_name}-{instance}-{guest}-{copy}-{network_index}"
                                '"]'
                            )
                    network_index = network_index + 1
        return resources

    def get_cr_resources_to_target_apply(self, instances):
        """
        Returns the name of the aws resource of the gsi_lab_aws module to target apply base on the instances number

        Parameters:
          instances (list(int)): instances to use

        Returns:
          list(str): resources name to target apply
        """

        resources = [
            "module.vpc",
            "aws_instance.student_access",
            "aws_security_group.bastion_host_sg",
            "aws_security_group.teacher_access_sg",
            "aws_key_pair.admin_pubkey",
            "aws_security_group.entry_point_sg",
        ]
        resources = resources + self._get_interface_resources_name(instances)
        resources = resources + self._get_subnet_resources_name(instances)
        resources = resources + self._get_security_group_resources_name(instances)
        resources = resources + self._get_machines_resources_name(instances)
        if self.description.teacher_access == "endpoint":
            resources.append("aws_ec2_instance_connect_endpoint.teacher_access[0]")
        else:
            resources.append("aws_instance.teacher_access_host[0]")
        if self.description.is_internet_access():
            resources.append("aws_route_table.scenario_internet_access[0]")
            resources.append("aws_security_group.internet_access_sg[0]")
            resources = resources + self._get_route_table_resources_name(instances)
        if self.description.configure_dns:
            resources = resources + self._get_dns_resources_name(instances)
        return resources

    def get_cr_resources_to_target_destroy(self, instances):
        """
        Returns the name of the aws resource of the gsi_lab_aws module to target destroy base on the instances number

        Parameters:
          instances (list(int)): instances to use

        Returns:
          list(str): resources name to target destroy
        """
        resources = self._get_machines_resources_name(instances)
        # resources = resources + self._get_interface_resources_name(instances) #TODO: fix this. If interfaces are added to be removed then all machines are removed.
        if self.description.configure_dns:
            resources = resources + self._get_dns_resources_name(instances)
        return resources

    def get_resources_to_recreate(self, instances, guests, copies):
        """
        Returns the name of the aws resource of the gsi_lab_aws module to recreate base on the machines names

        Parameters:
          instances (list(int)): instances to use

        Returns:
          list(str): resources name to recreate
        """
        machines_to_recreate = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy())
        student_access_name = (f"{self.description.institution}-{self.description.lab_name}-student_access")
        teacher_access_name = (f"{self.description.institution}-{self.description.lab_name}-teacher_access")

        resource_to_recreate = []
        if student_access_name in machines_to_recreate:
            machines_to_recreate.remove(student_access_name)
            resource_to_recreate.append("aws_instance.student_access")
        if teacher_access_name in machines_to_recreate:
            machines_to_recreate.remove(teacher_access_name)
            resource_to_recreate.append("aws_instance.teacher_access_host[0]")
        for machine in machines_to_recreate:
            resource_to_recreate.append('aws_instance.machines["' f"{machine}" '"]')
        return resource_to_recreate

    def can_delete_image(self, image_name):
        """
        Return true if the image is not being used by any machine.

        Parameters:
          image_name (str): name of the image

        Returns:
          bool: true if the image is not being used by any machine or false otherwise
        """
        image = self.client.get_image(image_name)
        if image:
            image_id = image[0]
            instances_images_ids = self.client.get_machines_imageid()
            for instance_image_id in instances_images_ids:
                if image_id == instance_image_id:
                    return False
        return True

    def list_instances(self, instances, guests, copies):
        machines_to_list = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy())
        headers = ["Name", "Status"]
        rows = []
        for machine in machines_to_list:
            machine_status = self.get_instance_status(machine)
            rows.append([machine,machine_status])
        return create_table(headers,rows)
    
    def shutdown(self, instances, guests, copies, stop_services):
        if stop_services:
            click.echo(f"Shutting down services...")
            for service in self.description.get_services_to_deploy():
                self.stop_instance(self.description.get_service_name(service))
        machines_to_shutdown = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy()) 
        for machine in machines_to_shutdown:
            click.echo(f"Shutting down instance {machine}...")
            self.stop_instance(machine)
            
    def start(self, instances, guests, copies, start_services):
        if start_services:
            click.echo(f"Booting up services...")
            for service in self.description.get_services_to_deploy():
                self.start_instance(self.description.get_service_name(service))
        machines_to_start = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy())  
        for machine in machines_to_start:
            click.echo(f"Booting up instance {machine}...")
            self.start_instance(machine)

    def reboot(self, instances, guests, copies, reboot_services):
        if reboot_services:
            click.echo(f"Booting up services...")
            for service in self.description.get_services_to_deploy():
                self.reboot_instance(self.description.get_service_name(service))
        machines_to_reboot = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy())
        for machine in machines_to_reboot:
            click.echo(f"Rebooting instance {machine}...")
            self.reboot_instance(machine)

    def recreate(self, instances, guests, copies, recreate_services=False):
        machines = self.description.parse_machines(instances, guests, copies, False, self.description.get_services_to_deploy())
        resources_to_recreate = self.get_resources_to_recreate(instances, guests, copies)
        click.echo("Recreating machines...")
        self.terraform_recreate(tectonic_resources.files('tectonic') / 'terraform' / 'modules' / 'gsi-lab-aws', resources_to_recreate)

        click.echo("Waiting for machines to boot up...")
        ansible = Ansible(self)
        ansible.wait_for_connections(instances, guests, copies, False, self.description.get_services_to_deploy())

        click.echo("Configuring student access...")
        entry_points = []
        for m in machines:
            base_name = self.description.get_base_name(m)
            if (base_name not in entry_points and
                (self.description.get_guest_attr(base_name, "entry_point", False) or
                base_name == "student_access")):
                entry_points.append(base_name)
        self._student_access(instances, entry_points)

        click.echo("Running after-clone configuration...")
        ansible.run(instances, guests, copies, quiet=True, only_instances=False)

        if self.description.deploy_elastic and self.description.monitor_type == "endpoint":
            click.echo("Configuring elastic agents...")
            self._elastic_install_endpoint(instances)

        if self.description.deploy_caldera:
            click.echo("Configuring caldera agents...")
            self._caldera_install_agent(instances)

    def student_access(self, instances):
        """Creates users for the students in the student access host
        and in all entry points of INSTANCES.

        Generates pseudo-random passwords and/or sets public SSH keys for the users.
        Returns a dictionary of created users.

        """
        entry_points = [ base_name for base_name, guest in self.description.guest_settings.items() if guest.get("entry_point") ]
        entry_points.append("student_access")
        self._student_access(instances, entry_points)

    def deploy_infraestructure(self, instances):
        """
        Deploy cyber range infrastructure
        """

        ansible = Ansible(self)
        click.echo("Deploying Cyber Range instances...")
        self._deploy_cr(tectonic_resources.files('tectonic') / 'terraform' / 'modules' / 'gsi-lab-aws',
                        self.get_deploy_cr_vars(),
                        instances)

        if len(self.description.get_services_to_deploy()) > 0: #Deploy services
            click.echo("Deploying Cyber Range services...")
            self._deploy_services(tectonic_resources.files('tectonic') / 'services' / 'terraform' / 'services-aws',
                                  self.get_deploy_services_vars(),
                                  instances)

            click.echo("Waiting for services to boot up...")
            services_to_deploy = []
            for service in self.description.get_services_to_deploy():
                services_to_deploy.append(self.description.get_service_name(service))
            extra_vars = {
                "elastic" : {
                    "monitor_type": self.description.monitor_type,
                    "deploy_policy": self.description.elastic_deploy_default_policy,
                    "policy_name": self.description.packetbeat_policy_name if self.description.monitor_type == "traffic" else self.description.endpoint_policy_name,
                    "http_proxy" : self.description.libvirt_proxy,
                    "description_path": self.description.description_dir,
                    "ip": self.client.get_machine_private_ip(self.description.get_service_name("elastic")),
                },
                "caldera":{
                    "ip": self.client.get_machine_private_ip(self.description.get_service_name("caldera")),
                    "description_path": self.description.description_dir,
                },
            }
            inventory = ansible.build_inventory(machine_list=services_to_deploy, extra_vars=extra_vars)
            ansible.wait_for_connections(inventory=inventory)

            click.echo("Configuring services...")
            ansible.run(inventory=inventory, playbook=self.ansible_services_path,quiet=True) 
            if self.description.deploy_elastic and self.description.monitor_type == "traffic":
                self._deploy_packetbeat()

        click.echo("Waiting for machines to boot up...")
        ansible.wait_for_connections(instances=instances)

        click.echo("Configuring student access...")
        self.student_access(instances)

        click.echo("Running after-clone configuration...")
        ansible.run(instances, quiet=True)

        if self.description.deploy_elastic and self.description.monitor_type == "endpoint":
            click.echo("Configuring elastic agents...")
            self._elastic_install_endpoint(instances)
        
        if self.description.deploy_caldera:
            click.echo("Configuring caldera agents...")
            self._caldera_install_agent(instances)

    def destroy_infraestructure(self, instances):
        """
        Destroy cyber range infrastructure
        """
        if len(self.description.get_services_to_deploy()) > 0: #Destroy services
            click.echo("Destroying Cyber Range services...")
            self._destroy_services(tectonic_resources.files('tectonic') / 'services' / 'terraform' / 'services-aws',
                                  self.get_deploy_services_vars(),
                                  instances)

        click.echo("Destroying Cyber Range instances...")
        self._destroy_cr(tectonic_resources.files('tectonic') / 'terraform' / 'modules' / 'gsi-lab-aws',
                         self.get_deploy_cr_vars(),
                         instances)

    def create_services_images(self, services):
        self.delete_services_images(services)
        super().create_services_images(services)
        # Delete temporary security groups created by Packer:
        self.client.delete_security_groups("Temporary group for Packer") 

    def delete_services_images(self, services):
        """
        Delete services base image.
        """
        for service in services:
            if services[service]:
                # Libvirt packer plugin fails if images exist.
                if not self.can_delete_image(service):
                    raise DeploymentAWSException(
                        f"Unable to delete image {service} because it is being used."
                    )
                self.client.delete_image(service)

    def create_cr_images(self, guests=None):
        self.delete_cr_images(guests)
        super().create_cr_images(guests)
        self.client.delete_security_groups("Temporary group for Packer")

    def get_deploy_services_vars(self):
        """Build the terraform variable dictionary for deployment of the CyberRange services."""
        return {
            "institution": self.description.institution,
            "lab_name": self.description.lab_name,
            "aws_region": self.description.aws_region,
            "network_cidr_block": self.description.network_cidr_block,
            "authorized_keys": self.description.authorized_keys,
            "subnets_json": json.dumps(self._get_services_network_data()),
            "guest_data_json": json.dumps(self._get_services_guest_data()),
            "os_data_json": json.dumps(OS_DATA),
            "configure_dns": self.description.configure_dns,
            "monitor_type": self.description.monitor_type,
            "packetbeat_vlan_id": self.description.packetbeat_vlan_id,
            "machines_to_monitor": self.description.get_machines_to_monitor(),
            "monitor": self.description.deploy_elastic,
        }
    
    def get_services_status(self):
        if len(self.description.get_services_to_deploy()) > 0:
            headers = ["Name", "Status"]
            rows = []
            if self.description.deploy_elastic:
                elastic_name = self.description.get_service_name("elastic")
                rows = [[elastic_name, self.get_instance_status(elastic_name)]]
                if self.description.monitor_type == "traffic":
                    packetbeat_name = self.description.get_service_name("packetbeat")
                    rows.append([packetbeat_name, self.get_instance_status(packetbeat_name)])
                else:
                    try:
                        if self.get_instance_status(elastic_name) == "RUNNING":
                            playbook = tectonic_resources.files('tectonic') / 'services' / 'elastic' / 'get_info.yml'
                            result = self._get_service_info("elastic",playbook,{"action":"agents_status"})
                            agents_status = result[0]['agents_status']
                            for key in agents_status:
                                rows.append([f"elastic-agents-{key}", agents_status[key]])
                        else:
                            click.echo(f"Unable to connect to Elastic. Check if machine is running.")
                    except Exception as e:
                        raise DeploymentAWSException(e)
                if self.description.deploy_caldera:
                    rows.append(["",""])
            if self.description.deploy_caldera:
                caldera_name = self.description.get_service_name("caldera")
                rows.append([caldera_name, self.get_instance_status(caldera_name)])
                try:
                    if self.get_instance_status(caldera_name) == "RUNNING":
                        playbook = tectonic_resources.files('tectonic') / 'services' / 'caldera' / 'get_info.yml'
                        result = self._get_service_info("caldera",playbook,{"action":"agents_status"})
                        response = result[0]['agents_status']
                        agents_status = {"alive": 0, "dead": 0, "pending_kill":0}
                        if len(response) > 0:
                            for agent in response: #TODO: see what the response is like when there are a large number of agents. pagination?
                                #Caldera uses this logic to define the state of the agent
                                now = int((datetime.datetime.utcnow() - datetime.datetime(1970, 1, 1)).total_seconds() * 1000) #Milliseconds since epoch
                                agent_last_seen = int((datetime.datetime.strptime(agent["last_seen"],"%Y-%m-%dT%H:%M:%SZ") - datetime.datetime(1970, 1, 1)).total_seconds() * 1000) #Milliseconds since epoch
                                difference = now - agent_last_seen
                                if (difference <= 60000 and agent["sleep_min"] == 3 and agent["sleep_max"] == 3 and agent["watchdog"] == 1):
                                    agents_status["pending_kill"] = agents_status["pending_kill"] + 1
                                elif (difference <= 60000 or difference <= (agent["sleep_max"] * 1000)):
                                    agents_status["alive"] = agents_status["alive"] + 1
                                else:
                                    agents_status["dead"] = agents_status["dead"] + 1
                        for key in agents_status:
                            rows.append([f"caldera-agents-{key}", agents_status[key]])
                    else:
                        click.echo(f"Unable to connect to Caldera. Check if machine is running.")
                except Exception as e:
                    raise DeploymentAWSException(e)
            return create_table(headers,rows)
        else:
            return "No services were deployed."
        
    def _get_services_network_data(self):
        """Compute the complete list of services subnetworks."""
        networks = {
            f"{self.description.institution}-{self.description.lab_name}-services" : {
                "cidr" : self.description.services_network,
                "mode": "none"
            },
        }
        return networks
    
    def _get_services_guest_data(self):
        """Compute the services guest data as expected by the deployment terraform module."""
        guest_data = {}
        if self.description.deploy_elastic:
            guest_data[self.description.get_service_name("elastic")] = {
                    "guest_name": self.description.get_service_name("elastic"),
                    "base_name": "elastic",
                    "hostname": "elastic",
                    "base_os": self.description.get_service_base_os("elastic"),
                    "internet_access": True,
                    "interfaces": {
                        f'{self.description.get_service_name("elastic")}-1' : {
                            "name": f'{self.description.get_service_name("elastic")}-1',
                            "index": 0,
                            "guest_name": self.description.get_service_name("elastic"),
                            "network_name": "services",
                            "subnetwork_name": f"{self.description.institution}-{self.description.lab_name}-services",
                            "private_ip": str(ipaddress.IPv4Network(self.description.services_network)[4]),
                            "mask": str(ipaddress.ip_network(self.description.services_network).prefixlen),
                        }
                    },
                    "disk": self.description.services["elastic"]["disk"],
                    "instance_type": self.description.instance_type.get_guest_instance_type(self.description.services["elastic"]["memory"],self.description.services["elastic"]["vcpu"],False,self.description.monitor_type),
                }
            if self.description.monitor_type == "traffic":
                guest_data[self.description.get_service_name("packetbeat")] = {
                    "guest_name": self.description.get_service_name("packetbeat"),
                    "base_name": "packetbeat",
                    "hostname": "packetbeat",
                    "base_os": self.description.get_service_base_os("packetbeat"),
                    "internet_access": False,
                    "interfaces": {
                        f'{self.description.get_service_name("packetbeat")}-1' : {
                            "name": f'{self.description.get_service_name("packetbeat")}-1',
                            "index": 0,
                            "guest_name": self.description.get_service_name("packetbeat"),
                            "network_name": "services",
                            "subnetwork_name": f"{self.description.institution}-{self.description.lab_name}-services",
                            "private_ip": str(ipaddress.IPv4Network(self.description.services_network)[5]),
                            "mask": str(ipaddress.ip_network(self.description.services_network).prefixlen),
                        },
                    },
                    "disk": self.description.services["packetbeat"]["disk"],
                    "instance_type": "t2.micro",
                }
        if self.description.deploy_caldera:
            guest_data[self.description.get_service_name("caldera")] = {
                    "guest_name": self.description.get_service_name("caldera"),
                    "base_name": "caldera", 
                    "hostname": "caldera",
                    "base_os": self.description.get_service_base_os("caldera"),
                    "internet_access": False,
                    "interfaces": {
                        f'{self.description.get_service_name("caldera")}-1' : {
                            "name": f'{self.description.get_service_name("caldera")}-1',
                            "index": 0,
                            "guest_name": self.description.get_service_name("caldera"),
                            "network_name": "services",
                            "subnetwork_name": f"{self.description.institution}-{self.description.lab_name}-services",
                            "private_ip": str(ipaddress.IPv4Network(self.description.services_network)[6]),
                            "mask": str(ipaddress.ip_network(self.description.services_network).prefixlen),
                        }
                    },
                    "disk": self.description.services["caldera"]["disk"],
                    "instance_type": self.description.instance_type.get_guest_instance_type(self.description.services["caldera"]["memory"],self.description.services["caldera"]["vcpu"],False,self.description.monitor_type),
                }
        return guest_data
        
    def _deploy_packetbeat(self):
        try:
            elastic_name = self.description.get_service_name("elastic")
            if self.get_instance_status(elastic_name) == "RUNNING":
                elastic_ip = self.get_ssh_hostname(elastic_name)
                playbook = tectonic_resources.files('tectonic') / 'services' / 'elastic' / 'get_info.yml'
                result = self._get_service_info("elastic",playbook,{"action":"get_token_by_policy_name","policy_name":self.description.packetbeat_policy_name})
                agent_token = result[0]["token"]
                ansible = Ansible(deployment=self)
                inventory = ansible.build_inventory(
                    machine_list=[self.description.get_service_name("packetbeat")],
                    username=OS_DATA[self.description.get_service_base_os("packetbeat")]["username"],
                    extra_vars = {
                        "action": "install",
                        "elastic_url": f"https://{elastic_ip}:8220",
                        "token": agent_token,
                        "elastic_agent_version": self.description.elastic_stack_version,
                        "institution": self.description.institution,
                        "lab_name": self.description.lab_name,
                    },
                )
                ansible.wait_for_connections(inventory=inventory)
                ansible.run(inventory = inventory,
                            playbook = tectonic_resources.files('tectonic') / 'services' / 'elastic' / 'agent_manage.yml',
                            quiet = True)
            else:
                click.echo(f"Unable to connect to Elastic. Check if machine is running.")
        except Exception as e:
            raise DeploymentAWSException(e)
    
    def get_services_resources_to_target_destroy(self, instances):
        resources = []
        if self.description.deploy_elastic and self.description.monitor_type == "traffic":
            resources = resources + self._get_sessions_resources_name(instances)
        return resources
        
    def get_services_resources_to_target_apply(self, instances):
        """
        Returns the name of the aws resource of the services-aws module to target apply base on the instances number

        Parameters:
          instances (list(int)): instances to use

        Returns:
          list(str): resources name to target apply
        """
        resources = [
            "aws_security_group.services_internet_access_sg[0]",
            "aws_security_group.caldera_scenario_sg",
            "aws_security_group.elastic_endpoint_scenario_sg",
            "aws_security_group.elastic_traffic_scenario_sg",
        ]
        resources = resources + self._get_services_instances_name()
        resources = resources + self._get_services_security_group_resources_name()
        resources = resources + self._get_services_interface_resources_name()
        if self.description.configure_dns:
            resources = resources + self._get_services_dns_resources_name()
        if self.description.deploy_elastic and self.description.monitor_type == "traffic":
            resources = resources + [
                "aws_ec2_traffic_mirror_target.packetbeat[0]",
                "aws_ec2_traffic_mirror_filter.filter[0]",
                "aws_ec2_traffic_mirror_filter_rule.filter_all_inbound[0]",
                "aws_ec2_traffic_mirror_filter_rule.filter_all_outbound[0]",
            ]
            resources = resources + self._get_sessions_resources_name(instances)
        return resources




# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

from tectonic.instance_type import InstanceType

class InstanceTypeAWS(InstanceType):

    instance_type_details = {
        "t2": [
            {"instance_type": "t2.nano", "vcpus": 1, "memory": 512},
            {"instance_type": "t2.micro", "vcpus": 1, "memory": 1024},
            {"instance_type": "t2.small", "vcpus": 1, "memory": 2048},
            {"instance_type": "t2.medium", "vcpus": 2, "memory": 4096},
            {"instance_type": "t2.large", "vcpus": 2, "memory": 8192},
            {"instance_type": "t2.xlarge", "vcpus": 4, "memory": 16384},
            {"instance_type": "t2.2xlarge", "vcpus": 8, "memory": 32768},
        ],
        "t3": [
            {"instance_type": "t3.nano", "vcpus": 2, "memory": 512},
            {"instance_type": "t3.micro", "vcpus": 2, "memory": 1024},
            {"instance_type": "t3.small", "vcpus": 2, "memory": 2048},
            {"instance_type": "t3.medium", "vcpus": 2, "memory": 4096},
            {"instance_type": "t3.large", "vcpus": 2, "memory": 8192},
            {"instance_type": "t3.xlarge", "vcpus": 4, "memory": 16384},
            {"instance_type": "t3.2xlarge", "vcpus": 8, "memory": 32768},
        ],
    }

    def get_guest_instance_type(self, memory, vcpus, monitor, monitor_type):
        """Returns a big enough AWS instance type for the given memory
        and cpus.

        A t2 instance will be used, unless monitor traffic is true, in
        which case a t3 instance will be used.

        If no suitable instance type can be used, None is returned.

        """
        if (
            not memory
            and not vcpus
        ):
            if monitor and monitor_type == "traffic":
                return self.default_instance_type.replace("t2.", "t3.")
            else:
                return self.default_instance_type

        if monitor and monitor_type == "traffic":
            instances = self.instance_type_details["t3"]
            selected_type = "t3"
        else:
            instances = self.instance_type_details["t2"]
            selected_type = "t2"

        for i in instances:
            if memory is None or memory <= i.get("memory"):
                if vcpus is None or vcpus <= i.get("vcpus"):
                    return i.get("instance_type")
        return f"{selected_type}.2xlarge"




# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

"""Constants used by Tectonic."""

OS_DATA = {
    "ubuntu22": {
        "ami_filter": "ubuntu/images/*/ubuntu-*-22.04-amd64-server*",
        "owner": "099720109477",
        "username": "ubuntu",
        "cloud_image_url": "https://cloud-images.ubuntu.com/releases/22.04/release-20230616/ubuntu-22.04-server-cloudimg-amd64.img",
        "cloud_image_checksum": "fe102bfb3d3d917d31068dd9a4bd8fcaeb1f529edda86783f8524fdc1477ee29",
    },
    "rocky8": {
        "ami_filter": "Rocky*8.8*x86_64",
        "owner": "792107900819",
        "username": "rocky",
        #"cloud_image_url": "http://dl.rockylinux.org/pub/rocky/8/images/x86_64/Rocky-8-GenericCloud.latest.x86_64.qcow2",
        "cloud_image_url": "file:///data/isos/Rocky-8-GenericCloud.latest.x86_64.qcow2",
        "cloud_image_checksum": "d17f15a7649dd064795306c114b90fc5062e7d5fefa9e9f0bd6b7ce1aeac3ae5",
    },
    "kali": {
        "ami_filter": "kali-last-snapshot-amd64*",
        "owner": "679593333241",  # TODO: Marketplace owner id
        "username": "kali",
        # cloud-init image not provided by Kali :(
        # See https://codingpackets.com/blog/kali-linux-cloud-init-image/
            "cloud_image_url": "file:///data/isos/kali-linux-2023.3-qemu-amd64.qcow2",
        "cloud_image_checksum": "b48cc396cf91ea1f38de4dca7abfcc0175f07d7af1b7e6df89a76194e7fda3d9",
        # "cloud_image_url": "https://cdimage.kali.org/kali-2023.3/kali-linux-2023.3-qemu-amd64.7z"
        # "cloud_image_checksum": "9ebbea4abb545c8e4a56153e2b5fa0ad90658c583af2fa76ad8c0a0b4ba23e20",
        },
    "windows_srv_2022": {
        "ami_filter": "Windows_Server-2022-English-Full-Base*",
        "owner": "amazon",
        "username": "administrator",
    },
}

ELASTICCLOUD_PRIVATELINK_MAPPING = {
    "af-south-1": {
        "vpc_service_name": "com.amazonaws.vpce.af-south-1.vpce-svc-0d3d7b74f60a6c32c",
        "private_domain_name": "vpce.af-south-1.aws.elastic-cloud.com",
        "az_names": ["af-south-1a", "af-south-1b", "af-south-1c"],
    },
    "ap-east-1": {
        "vpc_service_name": "com.amazonaws.vpce.ap-east-1.vpce-svc-0f96fbfaf55558d5c",
        "private_domain_name": "vpce.ap-east-1.aws.elastic-cloud.com",
        "az_names": ["ap-east-1a", "ap-east-1b", "ap-east-1c"],
    },
    "ap-northeast-1": {
        "vpc_service_name": "com.amazonaws.vpce.ap-northeast-1.vpce-svc-0e1046d7b48d5cf5f",
        "private_domain_name": "vpce.ap-northeast-1.aws.elastic-cloud.com",
        "az_names": ["ap-northeast-1b", "ap-northeast-1c", "ap-northeast-1d"],
    },
    "ap-northeast-2": {
        "vpc_service_name": "com.amazonaws.vpce.ap-northeast-2.vpce-svc-0d90cf62dae682b84",
        "private_domain_name": "vpce.ap-northeast-2.aws.elastic-cloud.com",
        "az_names": ["ap-northeast-2a", "ap-northeast-2b", "ap-northeast-2c"],
    },
    "ap-south-1": {
        "vpc_service_name": "com.amazonaws.vpce.ap-south-1.vpce-svc-0e9c1ae5caa269d1b",
        "private_domain_name": "vpce.ap-south-1.aws.elastic-cloud.com",
        "az_names": ["ap-south-1a", "ap-south-1b", "ap-south-1c"],
    },
    "ap-southeast-1": {
        "vpc_service_name": "com.amazonaws.vpce.ap-southeast-1.vpce-svc-0cbc6cb9bdb683a95",
        "private_domain_name": "vpce.ap-southeast-1.aws.elastic-cloud.com",
        "az_names": ["ap-southeast-1a", "ap-southeast-1b", "ap-southeast-1c"],
    },
    "ap-southeast-2": {
        "vpc_service_name": "com.amazonaws.vpce.ap-southeast-2.vpce-svc-0cde7432c1436ef13",
        "private_domain_name": "vpce.ap-southeast-2.aws.elastic-cloud.com",
        "az_names": ["ap-southeast-2a", "ap-southeast-2b", "ap-northeast-2c"],
    },
    "ca-central-1": {
        "vpc_service_name": "com.amazonaws.vpce.ca-central-1.vpce-svc-0d3e69dd6dd336c28",
        "private_domain_name": "vpce.ca-central-1.aws.elastic-cloud.com",
        "az_names": ["ca-central-1a", "ca-central-1b", "ca-central-1d"],
    },
    "eu-central-1": {
        "vpc_service_name": "com.amazonaws.vpce.eu-central-1.vpce-svc-081b2960e915a0861",
        "private_domain_name": "vpce.eu-central-1.aws.elastic-cloud.com",
        "az_names": ["eu-central-1a", "eu-central-1b", "eu-central-1c"],
    },
    "eu-south-1": {
        "vpc_service_name": "com.amazonaws.vpce.eu-south-1.vpce-svc-03d8fc8a66a755237",
        "private_domain_name": "vpce.eu-south-1.aws.elastic-cloud.com",
        "az_names": ["eu-south-1a", "eu-south-1b", "eu-south-1c"],
    },
    "eu-west-1": {
        "vpc_service_name": "com.amazonaws.vpce.eu-west-1.vpce-svc-01f2afe87944eb12b",
        "private_domain_name": "vpce.eu-west-1.aws.elastic-cloud.com",
        "az_names": ["eu-west-1a", "eu-west-1b", "eu-west-1c"],
    },
    "eu-west-2": {
        "vpc_service_name": "com.amazonaws.vpce.eu-west-2.vpce-svc-0e42a2c194c97a1d0",
        "private_domain_name": "vpce.eu-west-2.aws.elastic-cloud.com",
        "az_names": ["eu-west-2a", "eu-west-2b", "eu-west-2c"],
    },
    "eu-west-3": {
        "vpc_service_name": "com.amazonaws.vpce.eu-west-3.vpce-svc-0d6912d10db9693d1",
        "private_domain_name": "vpce.eu-west-3.aws.elastic-cloud.com",
        "az_names": ["eu-west-3a", "eu-west-3b", "eu-west-3c"],
    },
    "me-south-1": {
        "vpc_service_name": "com.amazonaws.vpce.me-south-1.vpce-svc-0381de3eb670dcb48",
        "private_domain_name": "vpce.me-south-1.aws.elastic-cloud.com",
        "az_names": ["me-south-3a", "me-south-3b", "me-south-3c"],
    },
    "sa-east-1": {
        "vpc_service_name": "com.amazonaws.vpce.sa-east-1.vpce-svc-0b2dbce7e04dae763",
        "private_domain_name": "vpce.sa-east-1.aws.elastic-cloud.com",
        "az_names": ["sa-east-1a", "sa-east-1b", "sa-east-1c"],
    },
    "us-east-1": {
        "vpc_service_name": "com.amazonaws.vpce.us-east-1.vpce-svc-0e42e1e06ed010238",
        "private_domain_name": "vpce.us-east-1.aws.elastic-cloud.com",
        "az_names": ["us-east-1a", "us-east-1b", "us-east-1e"],
    },
    "us-east-2": {
        "vpc_service_name": "com.amazonaws.vpce.us-east-2.vpce-svc-02d187d2849ffb478",
        "private_domain_name": "vpce.us-east-2.aws.elastic-cloud.com",
        "az_names": ["us-east-2a", "us-east-2b", "us-east-2c"],
    },
    "us-west-1": {
        "vpc_service_name": "com.amazonaws.vpce.us-west-1.vpce-svc-00def4a16a26cb1b4",
        "private_domain_name": "vpce.us-west-1.aws.elastic-cloud.com",
        "az_names": ["us-west-1a", "us-west-1b", "us-west-1c"],
    },
    "us-west-2": {
        "vpc_service_name": "com.amazonaws.vpce.us-west-2.vpce-svc-0e69febae1fb91870",
        "private_domain_name": "vpce.us-west-2.aws.elastic-cloud.com",
        "az_names": ["us-west-2a", "us-west-2b", "us-west-2c"],
    },
}

ELASTIC_SERVICE_BASE_URL = "https://api.elastic-cloud.com/api/v1"


SERVICES_SIZE = {
    "elastic" : {
        "libvirt" : {
            "test" : {
                "vcpu":2,
                "memory":4096, #MB
                "disk":10, #GB
            },
            "small" : {
                "vcpu":4,
                "memory":8192,
                "disk":60,
            },
            "medium": {
                "vcpu":8,
                "memory":16384,
                "disk":110,
            },
            "big":{
                "vcpu":12,
                "memory":24576,
                "disk":160,
            }
        },
        "aws" : {
            "small": "t2.small",
            "medium": "t2.medium",
            "big": "t2.big",
        }
    },
    "caldera" : {
        "libvirt" : {
            "test" : {
                "vcpu":1,
                "memory":2048, #MB
                "disk":10, #GB
            },
            "small" : {
                "vcpu":2,
                "memory":2048,
                "disk":20,
            },
            "medium": {
                "vcpu":3,
                "memory":4096,
                "disk":20,
            },
            "big":{
                "vcpu":4,
                "memory":8192,
                "disk":50,
            }
        },
        "aws" : {
            "small": "t2.small",
            "medium": "t2.medium",
            "big": "t2.big",
        }
    }
}



# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import re
import fabric
from paramiko import proxy
import logging

def ssh_version(path="/usr/bin/ssh"):
    """
    Return ssh client version.

    Parameters:
        path (str): path to ssh client binary.

    Returns:
        int: ssh client version.
    """
    try:
        p = subprocess.run([path, "-V"], check=False, capture_output=True)
        m = re.search(r"^OpenSSH_(\d+).\d+[^,]*,", p.stderr.decode())
        if m:
            return int(m.group(1))
        else:
            return None
    except FileNotFoundError:
        return None


def interactive_shell(
    hostname,
    username,
    gateway=None,
):
    """
    Interactively connects to a host.

    Parameters:
        hostname (str): The name target host to connect to.
        username (str): The user on the target host to use for the connection.
        gateway (str, tuple, None):  Can be a string to use as proxy, or a pair of strings of the form: (proxy_hostname, proxy_user), or None.
    """
    gateway_connection = None
    if isinstance(gateway, str):
        gateway_connection = gateway
    elif isinstance(gateway, tuple):
        gateway_connection = fabric.Connection(host=gateway[0], user=gateway[1])

    connection = fabric.Connection(host=hostname, user=username, gateway=gateway_connection)
    connection.shell()